# SPDX-License-Identifier: GPL-3.0-or-later
# vim: noet:ts=4
# (c) Emmanuel Beffara, 2002--2006
"""
Metapost support for Rubber.

The module parses input files for dependencies and does some checkings on
Metapost's log files after the process. Is it enough?
"""

import os, os.path
import re

import logging

msg = logging.getLogger(__name__)
from rubber.util import _, prog_available
import rubber.depend
import rubber.converters.latex

metapost_logfile_limit = 1000000


def check(source, target, context):
    return prog_available('mpost')


re_input = re.compile(r"input\s+(?P<file>[^\s;]+)")
# This is very restrictive, and so is the parsing routine. FIXME?
re_mpext = re.compile("[0-9]+|mpx|log")
re_mpxerr = re.compile("% line (?P<line>[0-9]+) (?P<file>.*)$")


class MPLogCheck(rubber.converters.latex.LogCheck):
    """
    This class adapats the LogCheck class from the main program to the case of
    MetaPost log files, which are very similar to TeX log files.
    """

    def __init__(self, pwd):
        super(MPLogCheck, self).__init__()
        self.pwd = pwd

    def read(self, name):
        """
        The read() method in LogCheck checks that the log is produced by TeX,
        here we check that it is produced by MetaPost.
        """
        with open(name, encoding='utf_8', errors='replace') as file:
            line = file.readline()
            if not line or line.find("This is MetaPost,") == -1:
                return 1
            self.lines = file.readlines()
        return 0

    def continued(self, line):
        """
        Messages in Metapost logs are broken at 79 characters per line, except
        in some cases where there are lines of this length that are not
        continued...
        """
        if len(line) != 79:
            return 0
        return line[-3:] != "..."

    def get_errors(self):
        """
        Parse the Metapost log file for errors. The file has the same form as
        a TeX log file, so the parser for TeX logs is used. The special case
        is that of TeX errors in Metapost labels, which requires parsing
        another TeX log file.
        """
        for err in super(MPLogCheck, self).get_errors():
            if (err["kind"] != "error" or err["text"] != "Unable to make mpx file."):
                yield err
                continue

            # a TeX error was found: parse mpxerr.log

            # TODO: Remove pylint suppression.
            # pylint: disable=no-member
            log = rubber.converter.latex.LogCheck()
            # FIXME this path has no testcase.
            if not log.readlog(os.path.join(self.pwd, "mpxerr.log"), metapost_logfile_limit):
                yield err
                continue

            # read mpxerr.tex to read line numbers from it

            with open(os.path.join(self.pwd, "mpxerr.tex"), encoding='utf_8',
                      errors='replace') as tex_file:
                tex = tex_file.readlines()

            # get the name of the mpxNNN.tex source

            for line in log.lines:
                if line[:2] == "**":
                    tex_src = os.path.join(".", line[2:].strip())
                    break

            for err in log.get_errors():
                if tex_src != err["file"]:
                    # the error is not in a Metapost source
                    yield err
                    continue

                line = int(err["line"])
                for shift in range(1, line + 1):
                    tex_line = tex[line - shift].rstrip()
                    m = re_mpxerr.search(tex_line)
                    if m:
                        err["line"] = int(m.group("line")) + shift - 2
                        err["file"] = os.path.join(self.pwd, m.group("file"))
                        err["pkg"] = "TeX"
                        yield err
                        break

                if shift == line:
                    # the error is in some verbatimtex
                    yield err


class Dep(rubber.depend.Node):
    """
    This class represents dependency nodes for MetaPost figures. The __init__
    method simply creates one node for the figures and one leaf node for all
    sources.
    """

    def __init__(self, env, target, source):
        self.cmd_pwd = os.path.dirname(source)
        super().__init__()
        self.add_product(target)
        self.include(os.path.basename(source))
        msg.debug(_("%s is made from %s"), target, " ".join(self.sources))
        self.base = source[:-3]
        self.cmd = ["mpost", "\\batchmode;input %s" % os.path.basename(self.base)]
        path = ':'.join(env.path)
        self.penv = {
            "TEXINPUTS": "%s:%s" % (path, os.getenv("TEXINPUTS", "")),
            "MPINPUTS": "%s:%s" % (path, os.getenv("MPINPUTS", ""))
        }
        self.log = None

    def include(self, source):
        """
        This function tries to find a specified MetaPost source (currently all
        in the same directory), appends its actual name to the list, and
        parses it to find recursively included files.
        """
        file = os.path.normpath(os.path.join(self.cmd_pwd, source))
        if os.path.exists(file + ".mp"):
            file = file + ".mp"
        elif not os.path.exists(file):
            return
        self.add_source(file)
        # In case UnicodeDecodeError, at worst a .mp file will
        # not be found.
        with open(file, encoding='utf_8', errors='replace') as fd:
            for line in fd:
                m = re_input.search(line)
                if m:
                    self.include(m.group("file"))

    def run(self):
        """
        Run Metapost from the source file's directory, so that figures are put
        next to their source file.
        """
        msg.info(_("running Metapost on %s"), self.base + ".mp")
        if rubber.util.execute(self.cmd, env=self.penv, pwd=self.cmd_pwd) == 0:
            return True

        # This creates a log file that has the same aspect as TeX logs.

        self.log = MPLogCheck(self.cmd_pwd)
        if not self.log.readlog(self.base + ".log", metapost_logfile_limit):
            msg.error(_(
                "I can't read MetaPost's log file, this is wrong."))
            return False
        return not self.log.errors()

    def get_errors(self):
        """
        Report the errors from the last compilation.
        """
        return self.log.get_errors()

    def clean(self):
        """
        This method removes all the files that the Metapost compilation may
        have created. It is required because the compilation may produce more
        files than just the figures used by the document.
        """
        base = self.base + "."
        ln = len(base)
        dir = os.path.dirname(base)
        if dir == "":
            list = os.listdir(".")
        else:
            list = os.listdir(dir)
        for f in list:
            file = os.path.join(dir, f)
            if file[:ln] == base:
                ext = file[ln:]
                m = re_mpext.match(ext)
                if m and ext[m.end():] == "":
                    msg.info(_("removing %s"), file)
                    os.remove(file)


# The `files' dictionary associates dependency nodes to MetaPost sources. It
# is used to detect when several figures from the same source are included. It
# uses a global variable, and this is authentically BAD. Therefore it deserves
# a big fat huge FIXME. Graphics modules should proably be encapsuled in
# classes the same way as standard Rubber modules, that would help for this
# problem.

files = {}


def convert(source, target, context, env):
    if source in files:
        dep = files[source]
        dep.add_product(target)
    else:
        dep = Dep(env, target, source)
        files[source] = dep
    return dep



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Sebastian Reichel, 2012
"""
Dependency analysis for package 'ltxtable' in Rubber.
"""

import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        self.doc = document
        document.hook_macro('LTXtable', 'aa', self.hook_ltxtable)

    def hook_ltxtable(self, loc, width, file):
        # If the file name looks like it contains a control sequence or a macro
        # argument, forget about this \LTXtable.
        if file.find('\\') < 0 and file.find('#') < 0:
            self.doc.add_source(file)



# SPDX-License-Identifier: GPL-3.0-or-later
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):

        document.program = 'lambda'
        document.engine = 'Omega'



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2002--2006
"""
Extraction of bounding box information from gzipped PostScript figures.
"""

import logging

msg = logging.getLogger(__name__)
from rubber.util import _
import rubber.depend

import gzip
import re

re_bbox = re.compile(r"%[%\w]*BoundingBox:")


class Dep(rubber.depend.Node):

    def __init__(self, target, source):
        super().__init__()
        self.add_product(target)
        self.add_source(source)
        self.source = source
        self.target = target

    def run(self):
        """
        This method reads the source file (which is supposed to be a
        gzip-compressed PostScript file) until it finds a line that contains a
        bounding box indication. Then it creates the target file with this
        single line.
        """
        msg.info(_("extracting bounding box from %s") % self.source)
        # Binary mode causes no decoding errors, but makes
        # difficult in Python to split UTF-8 input in lines.

        # Usual TeX encodings and UTF-8 share the property
        # that we can search for an ASCII pattern, and avoid
        # data corruption if other bytes are copied unchanged,
        # whichever character they represent.

        # Any 8-bit character has the advantage over UTF-8
        # that it never causes UnicodeDecodeError.
        with gzip.open(self.source, mode='rt', encoding='latin_1') as source:
            for line in source:
                if re_bbox.match(line):
                    with open(self.target, 'w', encoding='latin_1') as target:
                        target.write(line)
                    return True
        msg.error(_("no bounding box was found in %s!") % self.source)
        return False


def convert(source, target, context, env):
    return Dep(target, source)



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2002--2006
"""
Dependency analysis for the xr package.

The xr package allows one to put references in one document to other
(external) LaTeX documents. It works by reading the external document's .aux
file, so this support package registers these files as dependencies.
"""

from rubber.util import _
import logging

msg = logging.getLogger(__name__)
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        self.doc = document
        document.hook_macro('externaldocument', 'oa', self.hook_externaldocument)

    def hook_externaldocument(self, loc, opt, name):
        aux = self.doc.env.find_file(name + '.aux')
        if aux:
            self.doc.add_source(aux)
            msg.debug(_(
                "dependency %s added for external references") % aux)
        else:
            msg.debug(_(
                "file %s.aux is required by xr package but not found") % name)



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2002--2006
# vim: noet:ts=4
"""
VTeX support for Rubber.

This module specifies that the VTeX/Free compiler should be used. This
includes using "vlatex" of "vlatexp" instead of "latex" and knowing that this
produces a PDF or PostScript file directly. The PDF version is used by
default, switching to PS is possible by using the module option "ps".
"""

import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        document.engine = 'VTeX'
        if opt == 'ps':
            document.program = 'vlatexp'
            document.register_post_processor(old_suffix='.ps', new_suffix='.ps')
        else:
            document.program = 'vlatex'
            document.register_post_processor(old_suffix='.pdf', new_suffix='.pdf')
        document.cmdline = ['-n1', '@latex', '%s']



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2002--2006
# vim: noet:ts=4
"""
Support for the `graphics' package in Rubber.

This package is destined de become something rather large, to support standard
schemes for generating the figures that are supposed to be included, among
other features.

Another feature to include in the future concerns the parsing of the commands
that customize the operation of the package, such as \\DeclareGraphicsRule.

Currently, only dependency analysis is provided. The command parsing is
incomplete, because \\includegraphics can have a complex format.
"""

import os, os.path
import re
import logging

msg = logging.getLogger(__name__)
import rubber.depend
from rubber.util import _
from rubber.util import parse_keyval
from rubber.tex import parse_string
import rubber.module_interface

# default suffixes for each device driver (taken from the .def files)

# For dvips and dvipdf we put the suffixes .bb instead of .gz because these
# are the files LaTeX actually looks for. The module `eps_gz' declares the
# gzipped files as dependencies for them and extracts the bounding box
# information.

drv_suffixes = {
    "dvipdf": ["", ".eps", ".ps", ".eps.bb", ".ps.bb", ".eps.Z"],
    "dvipdfm": ["", ".jpg", ".jpeg", ".pdf", ".png"],
    "dvips": ["", ".eps", ".ps", ".eps.bb", ".ps.bb", ".eps.Z"],
    "dvipsone": ["", ".eps", ".ps", ".pcx", ".bmp"],
    "dviwin": ["", ".eps", ".ps", ".wmf", ".tif"],
    "emtex": ["", ".eps", ".ps", ".pcx", ".bmp"],
    "pctex32": ["", ".eps", ".ps", ".wmf", ".bmp"],
    "pctexhp": ["", ".pcl"],
    "pctexps": ["", ".eps", ".ps"],
    "pctexwin": ["", ".eps", ".ps", ".wmf", ".bmp"],
    "pdftex": ["", ".png", ".pdf", ".jpg", ".mps", ".tif"],
    "lualatex": ["", ".png", ".pdf", ".jpg", ".mps", ".tif"],
    "tcidvi": [""],
    "textures": ["", ".ps", ".eps", ".pict"],
    "truetex": ["", ".eps", ".ps"],
    "vtex": [
        "", ".gif", ".png", ".jpg", ".tif", ".bmp", ".tga", ".pcx", ".eps", ".ps", ".mps", ".emf",
        ".wmf"
    ]
}


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        self.doc = document
        document.hook_macro('includegraphics', '*oa', self.hook_includegraphics)
        document.hook_macro('graphicspath', 'a', self.hook_graphicspath)
        document.hook_macro('DeclareGraphicsExtensions', 'a', self.hook_declareExtensions)
        document.hook_macro('DeclareGraphicsRule', 'aaaa', self.hook_declareRule)

        self.prefixes = [os.path.join(x, '') for x in document.env.path]
        self.files = []

        #Latex accepts upper and lowercase filename extensions
        # to keep the above lists clean we auto-generate the
        # uppercase versions of the extensions automatically
        for engine, suffixes in drv_suffixes.items():
            suffixes += [x.upper() for x in suffixes]

        # I take dvips as the default, but it is not portable.
        if document.engine == 'pdfTeX' \
           and document.primary_product ().endswith ('.pdf'):
            self.suffixes = drv_suffixes['pdftex']
        elif document.engine == 'VTeX':
            self.suffixes = drv_suffixes['vtex']
        elif document.engine == 'LuaLaTeX':
            self.suffixes = drv_suffixes['lualatex']
        else:
            self.suffixes = drv_suffixes['dvips']

        # If the package was loaded with an option that matches the name of a
        # driver, use that driver instead.

        opts = parse_keyval(opt)

        for opt in opts.keys():
            if opt in drv_suffixes:
                self.suffixes = drv_suffixes[opt]

        document.env.graphics_suffixes = self.suffixes

    # Supported macros

    def hook_includegraphics(self, loc, starred, optional, name):
        # no suffixes are tried when the extension is explicit

        options = parse_keyval(optional)
        allowed_suffixes = self.suffixes

        # work out the file extension if given explicitly
        if 'ext' in options and options['ext']:
            # \includegraphics[ext=.png]{foo} and \includegraphics[ext=.png]{{{foo}}}
            allowed_suffixes = ['']
            if name.startswith('{{') and name.endswith('}}'):
                name = name[2:-2]
            name = name + options['ext']
        elif name.startswith('{{') and name.endswith('}}'):
            # \includegraphics{{{foo}}}
            name = name[2:-2]
        else:
            for suffix in self.suffixes:
                # if extension is given, includegraphics will always search for that
                # and disable automatic discovery
                if suffix and name.endswith(suffix):
                    allowed_suffixes = ['']
                    # support special syntax \includegraphics{subdir/{foo.1}.png}
                    # at least some of the time
                    if name.endswith('}' + suffix):
                        # try to emulate what the TeX parser does.
                        name = re.sub('{([^{]*)}', '\\1', name)
                    break

        # If the file name looks like it contains a control sequence or a macro
        # argument, forget about this \includegraphics.

        if name.find('\\') >= 0 or name.find('#') >= 0:
            return

        # We only accept conversions from file types we don't know and cannot
        # produce.

        def check(vars):
            source = vars['source']
            if os.path.exists(vars['target']) and self.doc.env.may_produce(source):
                return False
            if self.suffixes == ['']:
                return True
            for suffix in allowed_suffixes:
                if source[-len(suffix):] == suffix:
                    return False
            return True

        node = self.doc.env.convert(name,
                                    suffixes=allowed_suffixes,
                                    prefixes=self.prefixes,
                                    check=check,
                                    context=self.doc.vars)

        if isinstance(node, str):
            msg.debug(_("graphics %s found in %s"), name, node)
            self.doc.add_source(node)
        elif isinstance(node, rubber.depend.Node):
            msg.debug(_("graphics %s converted from %s"),
                      name, node.primary_product())
            self.doc.add_source(node.primary_product())
            self.files.append(node)
        else:
            assert node is None
            msg.warning(rubber.util._format(loc, _("graphics `%s' not found") % name))

    def hook_graphicspath(self, loc, arg):
        # The argument of \graphicspath is a list (in the sense of TeX) of
        # prefixes that can be put in front of graphics names.
        parser = parse_string(arg)
        while True:
            arg = parser.get_argument_text()
            if arg is None:
                break
            self.prefixes.insert(0, arg)

    def hook_declareExtensions(self, loc, list):
        for suffix in list.split(","):
            self.suffixes.insert(0, suffix.strip())

    def hook_declareRule(self, loc, ext, type, read, command):
        if read in self.suffixes:
            return
        self.suffixes.insert(0, read)
        msg.debug("*** FIXME ***  rule %s -> %s [%s]" % (ext.strip(), read, type))

    #  module interface

    def pre_compile(self):
        # Pre-compilation means running all needed conversions. This is not done
        # through the standard dependency mechanism because we do not want to
        # interrupt compilation when a graphic is not found.
        for node in self.files:
            assert not node.making
            node.make()
        return True



# SPDX-License-Identifier: GPL-3.0-or-later
# Sebastian Kapfer <sebastian.kapfer@fau.de> 2015.
# based on code by Sebastian Reichel and others.
# vim: noet:ts=4
"""
Bibliographies (Biber and BibTeX).
"""

from rubber.util import _
import logging

msg = logging.getLogger(__name__)
import rubber.util
import rubber.depend
import os
import re
import subprocess
from abc import ABC, abstractmethod


class BibToolDep(rubber.depend.Node, ABC):

    def __init__(self):
        super().__init__()
        self.tool = "bibtex"
        self.environ = os.environ.copy()
        self.bib_paths = rubber.util.explode_path("BIBINPUTS")
        self.bst_paths = rubber.util.explode_path("BSTINPUTS")
        self.blg = None

    @abstractmethod
    def build_command(self):
        pass

    def do_path(self, args):
        if len(args) != 1:
            raise SyntaxError(_("invalid syntax for directive '%s'") % "path")
        path = args[0]
        self.bib_paths.insert(0, path)

    def run(self):
        # command might have been updated in the mean time, so get it now
        self.environ["BIBINPUTS"] = ":".join(self.bib_paths)
        self.environ["BSTINPUTS"] = ":".join(self.bst_paths)
        command = self.build_command()

        msg.info(_("running: %s") % " ".join(command))
        process = subprocess.Popen(command,
                                   stdin=subprocess.DEVNULL,
                                   stdout=subprocess.DEVNULL,
                                   env=self.environ)
        if process.wait() != 0:
            msg.error(_("There were errors running %s.") % self.tool)
            return False
        return True

    def find_bib(self, name):
        return rubber.util.find_resource(name, suffix=".bib", paths=self.bib_paths)

    def get_errors(self):
        """
        Read the log file, identify error messages and report them.
        """
        try:
            log = open(self.blg, encoding='utf_8', errors='replace')
        except:
            msg.warning(_("cannot open BibTeX logfile: %s") % self.blg)
            return

        with log:
            last_line = ""
            for line in log:
                m = re_error.search(line)
                if m:
                    # TODO: it would be possible to report the offending code.
                    if m.start() == 0:
                        text = last_line.strip()
                    else:
                        text = line[:m.start()].strip()

                    # when including a bibtex DB with \bibliography{a.bib}
                    # bibtex will report it in the log as a.bib.bib.
                    # work around this
                    filename = m.group("file")
                    if filename.endswith(".bib.bib"):
                        filename = filename[:-4]

                    filename = self.find_bib(filename) or filename

                    d = {"pkg": "bibtex", "kind": "error", "file": filename, "text": text}

                    if m.group("line"):
                        d["line"] = int(m.group("line"))

                    yield d

                last_line = line


# The regular expression that identifies errors in BibTeX log files is heavily
# heuristic. The remark is that all error messages end with a text of the form
# "---line xxx of file yyy" or "---while reading file zzz". The actual error
# is either the text before the dashes or the text on the previous line.

re_error = re.compile("---(line (?P<line>[0-9]+) of|while reading) file (?P<file>.*)")


class BibTeXDep(BibToolDep):
    """
    This class represents a single bibliography for a document.
    """

    def __init__(self, document, aux_basename):
        """
        Initialise the bibiliography for the given document. The base name is
        that of the aux file from which citations are taken.
        """
        super().__init__()

        self.log = document.basename(with_suffix=".log")
        self.aux = aux_basename + ".aux"
        self.bbl = aux_basename + ".bbl"
        self.blg = aux_basename + ".blg"
        self.add_product(self.bbl)
        self.add_product(self.blg)
        self.add_source(self.aux)
        document.add_source(self.bbl)

        self.bst_file = None
        self.set_style("plain")
        self.db = {}
        self.crossrefs = None

    def build_command(self):
        ret = [self.tool]
        if self.crossrefs is not None:
            ret.append("-min-crossrefs=" + self.crossrefs)
        ret.append(self.aux)
        return ret

    #
    # The following method are used to specify the various datafiles that
    # BibTeX uses.
    #
    def do_crossrefs(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '{}'")
                                     .format('crossrefs'))
        number = args[0]
        self.crossrefs = number

    def do_stylepath(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '{}'")
                                     .format('stylepath'))
        path = args[0]
        self.bst_paths.insert(0, path)

    def do_sorted(self, args):
        msg.debug(_("directive '%s' is no longer supported") % "sorted")

    def do_tool(self, args):
        if len(args) != 1:
            raise rubber.SyntaxError(_("invalid syntax for directive '{}'")
                                     .format('tool'))
        tool = args[0]
        # FIXME document this in user documentation
        self.tool = tool

    def hook_bibliography(self, loc, bibs):
        for name in bibs.split(","):
            filename = self.find_bib(name)
            if filename is not None:
                self.db[name] = filename
                self.add_source(filename)
            else:
                msg.error(_("cannot find bibliography resource %s") % name)

    def hook_bibliographystyle(self, loc, name):
        """
        Define the bibliography style used. This method is called when
        \\bibliographystyle is found. If the style file is found in the
        current directory, it is considered a dependency.
        """
        self.set_style(name)

    def set_style(self, name):
        if self.bst_file is not None:
            self.remove_source(self.bst_file)
            self.bst_file = None

        filename = rubber.util.find_resource(name, suffix=".bst", paths=self.bst_paths)
        if filename is not None:
            self.bst_file = filename
            self.add_source(filename)
        elif name not in ["plain", "alpha"]:
            # do not complain about default styles coming with bibtex
            msg.warning(_("cannot find bibliography style %s") % name)



# SPDX-License-Identifier: GPL-3.0-or-later
# module_interface.py, part of Rubber building system for LaTeX documents..

# Copyright (C) 2015-2015 Nicolas Boulenguez <nicolas.boulenguez@free.fr>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This module defines the common interface for all rubber modules.

Most of them are implemented as Python modules under
rubber.latex_modules.  Each one contains a subclass of
rubber.module_interface.Module that will be instantiated once when the
plugin is loaded by rubber.converters.latex.

rubber.converters.latex also declares a subclass that is instantiated
each time a .rub file is read.
"""

from abc import ABC

from rubber.util import _
import logging

msg = logging.getLogger(__name__)
import rubber.util


class Module(ABC):
    """
    This is the base class for modules. Each module should define a class
    named 'Module' that derives from this one. The default implementation
    provides all required methods with no effects.

    The constructor is mandatory. Its profile must be
    def __init__ (self, document, opt):

    'document' is the compiling environment (an instance of
    converters.latex.LaTeXDep)

    'opt' describes the option given to the macro that caused
    the module to load.
    """

    def pre_compile(self):
        """
        This method is called before the first LaTeX compilation. It is
        supposed to build any file that LaTeX would require to compile the
        document correctly. The method must return true on success.
        """
        return True

    def post_compile(self):
        """
        This method is called after each LaTeX compilation. It is supposed to
        process the compilation results and possibly request a new
        compilation. The method must return true on success.
        """
        return True

    def clean(self):
        """
        Remove additional files for this LaTeX module.
        Nothing recursive happens here.
        Files registered as products are removed by rubber.clean ().
        """

    def command(self, cmd, args, _dep=None):
        """
        This is called when a directive for the module is found in the source.
        We treat syntax errors in the directive as fatal, aborting the run.

        The _dep parameter is an implementation detail and must not
        be used outside the latex_modules subdirectory.
        When absent, self.dep is tried.
        """
        try:
            if _dep is None:
                _dep = self.dep  # may raise AttributeError
            handler = getattr(_dep, "do_" + cmd)  # may raise AttributeError
        except AttributeError:
            # there is no do_ method for this directive, which means there
            # is no such directive.
            raise rubber.SyntaxError(_("no such directive '%s' (in module %s)") % (cmd, self.__module__))
        handler(args)

    def get_errors(self):
        """
        This is called if something has failed during an operation performed
        by this module. The method returns a generator with items of the same
        form as in LaTeXDep.get_errors.
        """
        # TODO: what does this mean (copied from rubber.converters.latex)?
        if None:
            yield None



# SPDX-License-Identifier: GPL-3.0-or-later
import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):

        brf = document.basename(with_suffix='.brf')
        document.add_product(brf)
        document.watch_file(brf)



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2004--2006
# vim: noet:ts=4
"""
Multibib support for Rubber

This package allows several bibliographies in one document. Each occurence of
the \\newcites macro creates a new bibliography with its associated commands,
using a new aux file. This modules behaves like the default BibTeX module for
each of those files.

The directives are the same as those of the BibTeX module. They all accept an
optional argument first, enclosed in parentheses as in "multibib.path
(foo,bar) here/", to specify which bibliography they apply to. Without this
argument, they apply to all bibliographies.
"""

import re
from rubber.util import _
import logging

msg = logging.getLogger(__name__)
import rubber.biblio
import rubber.module_interface

re_optarg = re.compile(r'\((?P<list>[^()]*)\) *')


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        self.doc = document
        self.bibs = {}
        self.defaults = []
        self.commands = {}
        document.hook_macro('newcites', 'a', self.hook_newcites)

    def command(self, cmd, args):
        names = None

        # Check if there is the optional argument.

        if len(args) > 0:
            match = re_optarg.match(args[0])
            if match:
                names = match.group('list').split(',')
                args = args[1:]

        # If not, this command will also be executed for newly created indices
        # later on.

        if names is None:
            self.defaults.append([cmd, args])
            names = self.bibs.keys()

        # Then run the command for each index it concerns.

        for name in names:
            if name in self.bibs:
                super().command(cmd, args, _dep=self.bibs[name])
            elif name in self.commands:
                self.commands[name].append([cmd, args])
            else:
                self.commands[name] = [[cmd, args]]

    def hook_newcites(self, loc, name):
        self.doc.add_product(name + ".aux")
        bib = self.bibs[name] = rubber.biblio.BibTeXDep(self.doc, name)
        self.doc.hook_macro('bibliography' + name, 'a', bib.hook_bibliography)
        self.doc.hook_macro('bibliographystyle' + name, 'a', bib.hook_bibliographystyle)
        for cmd in self.defaults:
            super().command(*cmd, _dep=bib)
        if name in self.commands:
            for cmd in self.commands[name]:
                super().command(*cmd, _dep=bib)
        msg.debug(_("bibliography %s registered") % name)



# SPDX-License-Identifier: GPL-3.0-or-later
# (c) Emmanuel Beffara, 2003--2006
"""
Dependency analysis and environment parsing for package 'verbatim' in Rubber.
"""

import rubber.module_interface


class Module(rubber.module_interface.Module):

    def __init__(self, document, opt):
        self.doc = document
        document.hook_macro('verbatiminput', 'a', self.hook_input)
        document.hook_begin('comment', self.hook_begin_comment)

    def hook_begin_comment(self, loc):
        self.doc.h_begin_verbatim(loc, env='comment')

    def hook_input(self, loc, file):
        if file.find('\\') < 0 and file.find('#') < 0:
            self.doc.add_source(file)
